import java.util.Timer;
import java.util.TimerTask;
import com.nokia.mid.ui.FullCanvas;

import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.midlet.MIDletStateChangeException;
import javax.microedition.rms.RecordStore;

public class GameCanvas extends FullCanvas{

	//the block's time task: call onBlockTimer
	class BlockTimerTask extends TimerTask{

		private GameCanvas gameCanvas;
		
		public BlockTimerTask(GameCanvas gameCanvas){
			this.gameCanvas = gameCanvas;
		}
		
		public void run() {
			gameCanvas.onBlockTimer();
		}
	}
	
	//key settings
	private final static int KeyOK = FullCanvas.KEY_SOFTKEY1;
	private final static int KeyCancel = FullCanvas.KEY_SOFTKEY2;
	private final static int KeyCenter = FullCanvas.KEY_SOFTKEY3;
	private final static int KeyUp = FullCanvas.KEY_UP_ARROW;
	private final static int KeyDown = FullCanvas.KEY_DOWN_ARROW;
	private final static int KeyLeft = FullCanvas.KEY_LEFT_ARROW;
	private final static int KeyRight = FullCanvas.KEY_RIGHT_ARROW;
	
	//game status
	private final static int ShowSplash = 0;
	private final static int ShowMainMenu = 1;
	private final static int Gaming = 2;
	private final static int GamePause = 3;
	private final static int GameCompleted = 4;
	
	//colors
	public final static int BackgroundColor = 0x00000000;
	public final static int MenuBackgroundColor = 0x00252525;
	public final static int BlockDownTipsColor = 0x00404040;
	public final static int MenuTitleColor = 0x00752323;
	public final static int MenuNormalColor = 0x00668B8B;
	public final static int MenuHighlightColor = 0x0087CEFA;
	public final static int GridColor = 0x00303030;
	public final static int GridFrameColor = 0x00999999;
	public final static int SoftBottonColor = 0x008B8B66;
	public final static int GamingTextColor = 0x00CAE1FF;
	public final static int CurrentBlockBackgroundColor = GridColor;
	
	private QTetrisApp theApp;
	
	public final static int GridWidth = 10;
	public final static int GridHeight = 20;
	private final static int maxLevel = 10;
	private final static int BlockCountPerLevel = 30;
	
	private final int[][] grid = new int[GridWidth][GridHeight];
	private final int[][] gridColors = new int[GridWidth][GridHeight];

	private Timer blockTimer;
	private int level = 1;
	private long score = 0;
	public long highestScore = 0;
	
	private int status = ShowSplash;
	
	private Block currentBlock;
	private Block[] nextBlocks = new Block[4];
	private int passedBlockCount = 0;
	
	private int gridXStart;
	private int gridXEnd;
	private int gridYStart;
	private int gridYEnd;
	
	private int menuSelection = 0;

	public GameCanvas(QTetrisApp app) {
		theApp = app;
		gridXStart = 55;
		gridXEnd = getWidth() - 55;
		gridYStart = 20;
		gridYEnd = getHeight() - 40;
		status = ShowSplash;		
		readHighestScore();
	}

	private void readHighestScore(){
		RecordStore rs = null;
		try {
			rs = RecordStore.openRecordStore("QTetris", false);
			byte[] bytes = rs.getRecord(1);
			highestScore = (long)(
		            (long)(0xFF & bytes[0]) << 56  |
		            (long)(0xFF & bytes[1]) << 48  |
		            (long)(0xFF & bytes[2]) << 40  |
		            (long)(0xFF & bytes[3]) << 32  |
		            (long)(0xFF & bytes[4]) << 24  |
		            (long)(0xFF & bytes[5]) << 16  |
		            (long)(0xFF & bytes[6]) << 8   |
		            (long)(0xFF & bytes[7]) << 0
		            );
			rs.closeRecordStore();
		} catch (Exception e) {
			if (rs != null) {
				try {
					rs.closeRecordStore();
				} catch (Exception e2) {
				}
			}
		}
	}
	
	private void saveHighestScore(){
		RecordStore rs = null;
		try {
			rs = RecordStore.openRecordStore("QTetris", true);
			byte[] bytes = new byte[] {
			        (byte)((highestScore >> 56) & 0xFF),
			        (byte)((highestScore >> 48) & 0xFF),
			        (byte)((highestScore >> 40) & 0xFF),
			        (byte)((highestScore >> 32) & 0xFF),
			        (byte)((highestScore >> 24) & 0xFF),
			        (byte)((highestScore >> 16) & 0xFF),
			        (byte)((highestScore >> 8)  & 0xFF),
			        (byte)((highestScore >> 0)  & 0xFF),
			    };
			if (rs.getNumRecords() == 0)
				rs.addRecord(bytes, 0, bytes.length);
			else
				rs.setRecord(1, bytes, 0, bytes.length);
			rs.closeRecordStore();
		} catch (Exception e) {
			if (rs != null) {
				try {
					rs.closeRecordStore();
				} catch (Exception e2) {
				}
			}
		}
	}
	
	public void exit(){
		saveHighestScore();
		try {
			Display.getDisplay(theApp).setCurrent(null);
			theApp.destroyApp(false);
			theApp.notifyDestroyed();
		} catch (MIDletStateChangeException e) {}
	}
	
	private int getBlockTimerTicksByLevel(){
		//return (1000 - (1000 - rePaintTime)/maxLevel*level);
		return (1000 - (1000 - 50)/maxLevel*level);
	}
	
	private void startNewGame(){
		passedBlockCount = 0;
		score = 0;
		for(int i=0; i<GridWidth; i++)
			for(int j=0; j<GridHeight; j++)
				grid[i][j] = 0;
		currentBlock = Block.CreateRandomBlock();
		for(int i=0; i<nextBlocks.length; i++)
			nextBlocks[i] = Block.CreateRandomBlock();
		if(blockTimer != null) blockTimer.cancel();
		blockTimer = new Timer();
		blockTimer.scheduleAtFixedRate(new BlockTimerTask(this), 0, getBlockTimerTicksByLevel());
		status = Gaming;
	}
	
	private void pauseGame(){
		if(blockTimer != null) blockTimer.cancel();
		status = GamePause;
	}
	
	private void restoreGame(){
		if(blockTimer != null) blockTimer.cancel();
		blockTimer = new Timer();
		blockTimer.scheduleAtFixedRate(new BlockTimerTask(this), 0, getBlockTimerTicksByLevel());
		status = Gaming;
	}
	
	private void endGame(){
		if(blockTimer != null) blockTimer.cancel();
		status = GameCompleted;
	}

	protected synchronized void keyPressed(int keyCode) {
		if (status == ShowSplash) {
			if (keyCode == KeyCancel) 							exit();
			else if (keyCode == KeyOK || keyCode == KeyCenter) 	{status = ShowMainMenu; level = 1;}
		}
		else if (status == Gaming) {
			if (keyCode == KeyCancel) 							{pauseGame(); menuSelection = 0;}
			else if (keyCode == KeyLeft) 						currentBlock.moveLeft(grid);
			else if (keyCode == KeyRight) 						currentBlock.moveRight(grid);
			else if (keyCode == KeyUp) 							currentBlock.rotate(grid);
			else if (keyCode == KeyDown) 						{checkAndReduce(); currentBlock.moveDown(grid); score++;}
			else if (keyCode == KeyCenter) 						{score += currentBlock.moveBottom(grid); checkAndReduce();}
		}
		else if(status == GamePause){
			if(keyCode == KeyUp) 								menuSelection = (menuSelection + 3)%4;
			else if(keyCode == KeyDown) 						menuSelection = (menuSelection + 1)%4;
			else if(keyCode == KeyCancel) 						restoreGame();
			else if(keyCode == KeyOK || keyCode == KeyCenter){
				if(menuSelection == 0) 							restoreGame();
				else if(menuSelection == 1) 					endGame();
				else if(menuSelection == 2) 					{endGame(); status = ShowMainMenu; level = 1;}
				else 											exit();
			}
		}
		else if(status == GameCompleted){
			if(keyCode == KeyOK || keyCode == KeyCenter) 		{status = ShowMainMenu; level = 1;}
			else if(keyCode == KeyCancel) 						exit();
		}
		else if(status == ShowMainMenu){
			if(keyCode == KeyCancel) 							exit();
			else if(keyCode == KeyLeft) 						level = (level + maxLevel - 2)%maxLevel + 1;
			else if(keyCode == KeyRight) 						level = level%maxLevel + 1;
			else if(keyCode == KeyOK || keyCode == KeyCenter)  	startNewGame();
		}
		this.repaint();
	}

	protected synchronized void keyRepeated(int keyCode){
		if(status == Gaming &&
				(keyCode == KeyDown || keyCode == KeyLeft || keyCode == KeyRight))
			keyPressed(keyCode);
	}
	
	protected void paint(Graphics g) {
		if(status == ShowSplash) paintSplash(g);
		else if (status == Gaming) paintGaming(g);
		else if (status == GamePause) paintGamePause(g);
		else if (status == GameCompleted) paintGameCompleted(g);
		else if (status == ShowMainMenu) paintMainMenu(g);
		else paintConfution(g);
	}

	private void paintMainMenu(Graphics g) {
		g.setColor(MenuBackgroundColor);
		g.fillRect(0, 0, getWidth(), getHeight());
		
		g.setColor(MenuTitleColor);
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE));
		g.drawString("MAIN MENU", getWidth()/2, 10, Graphics.HCENTER | Graphics.TOP);

		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		g.setColor(MenuHighlightColor);
		g.drawString("SELECT LEVEL: <" + level + ">", getWidth()/2, 40, Graphics.HCENTER | Graphics.TOP);
		
		g.setColor(MenuNormalColor);
		String strScore = "HIGHEST SCORE: " + highestScore;
		g.drawString(strScore, getWidth()/2, 60, Graphics.HCENTER | Graphics.TOP);
		
		paintSoftMenu(g, "START NEW GAME", "EXIT");
	}

	private void paintGameCompleted(Graphics g) {
		g.setColor(MenuBackgroundColor);
		g.fillRect(0, 0, getWidth(), getHeight());
		
		g.setColor(MenuTitleColor);
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE));
		g.drawString("GAME END", getWidth()/2, 10, Graphics.HCENTER | Graphics.TOP);

		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		g.setColor(MenuNormalColor);
		g.drawString("LEVEL: " + level, getWidth()/2, 40, Graphics.HCENTER | Graphics.TOP);
		
		String strScore = "SCORE: " + score;
		boolean bNewHighestScore = (score > highestScore);
		if (bNewHighestScore) {
			highestScore = score;
			strScore = "NEW HIGHEST " + strScore;
		}
		g.drawString(strScore, getWidth() / 2, 60, Graphics.HCENTER | Graphics.TOP);
		if(bNewHighestScore == false)
			g.drawString("HIGHEST SCORE: " + highestScore, getWidth()/2, 80, Graphics.HCENTER | Graphics.TOP);
		
		paintSoftMenu(g, "BACK TO MENU", "EXIT");
	}

	private void paintGamePause(Graphics g) {
		g.setColor(MenuBackgroundColor);
		g.fillRect(gridXStart, gridYStart, gridXEnd-gridXStart+1, gridYEnd-gridYStart+1);
		
		g.setColor(MenuTitleColor);
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE));
		g.drawString("PAUSED", (gridXEnd+gridXStart)/2, gridYStart + 10, Graphics.HCENTER | Graphics.TOP);

		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		g.setColor(menuSelection <= 0 ? MenuHighlightColor : MenuNormalColor);
		g.drawString("RESTORE", (gridXEnd+gridXStart)/2, gridYStart + 40, Graphics.HCENTER | Graphics.TOP);
		g.setColor(menuSelection == 1 ? MenuHighlightColor : MenuNormalColor);
		g.drawString("END", (gridXEnd+gridXStart)/2, gridYStart + 60, Graphics.HCENTER | Graphics.TOP);
		g.setColor(menuSelection == 2 ? MenuHighlightColor : MenuNormalColor);
		g.drawString("MAIN MENU", (gridXEnd+gridXStart)/2, gridYStart + 80, Graphics.HCENTER | Graphics.TOP);
		g.setColor(menuSelection >= 3 ? MenuHighlightColor : MenuNormalColor);
		g.drawString("EXIT", (gridXEnd+gridXStart)/2, gridYStart + 100, Graphics.HCENTER | Graphics.TOP);

		g.setColor(BackgroundColor);
		g.fillRect(0, gridYEnd + 5, getWidth(), getHeight() - gridYEnd - 5);
		paintSoftMenu(g, "OK", "CANCEL");
	}

	private void paintConfution(Graphics g) {
		g.setColor(BackgroundColor);
		g.fillRect(0, 0, getWidth(), getHeight());
		g.setColor(MenuHighlightColor);
		g.drawString("Unimplemented content!", getWidth()/2, getHeight()/2, Graphics.HCENTER | Graphics.BASELINE);
	}

	private void paintGaming(Graphics g){
		//paint background
		g.setColor(BackgroundColor);
		g.fillRect(0, 0, getWidth(), getHeight());
		//paint grid content
		int columnWidth = (gridXEnd - gridXStart)/GridWidth;
		int rowHeight = (gridYEnd - gridYStart)/GridHeight;
		for(int i=0; i<GridWidth; i++){
			for(int j=0; j<GridHeight; j++){
				if(grid[i][j] <= 0) continue;
				int x = gridXStart + i * columnWidth;
				int y = gridYStart + j * rowHeight;
				g.setColor(gridColors[i][j]);
				g.fillRect(x, y, columnWidth, rowHeight);
			}
		}
		//paint shadow block
		Block shadowBlock = currentBlock.Clone();
		shadowBlock.Color = BlockDownTipsColor;
		shadowBlock.moveBottom(grid);
		shadowBlock.paintInGameCanvaGrid(g, gridXStart, gridYStart, columnWidth, rowHeight);
		//paint the current block
		currentBlock.paintInGameCanvaGrid(g, gridXStart, gridYStart, columnWidth, rowHeight);
		//paint grid frame
		g.setColor(GridColor);
		for(int i=0; i<=GridWidth; i++)
			g.drawLine(gridXStart + i*columnWidth, gridYStart, gridXStart + i*columnWidth, gridYEnd);
		for(int j=0; j<=GridHeight; j++)
			g.drawLine(gridXStart, gridYStart + j*rowHeight, gridXEnd, gridYStart + j*rowHeight);
		g.setColor(GridFrameColor);
		g.drawLine(gridXStart, gridYStart, gridXEnd, gridYStart);
		g.drawLine(gridXEnd, gridYStart, gridXEnd, gridYEnd);
		g.drawLine(gridXEnd, gridYEnd, gridXStart, gridYEnd);
		g.drawLine(gridXStart, gridYEnd, gridXStart, gridYStart);
		//paint level, score and count
		g.setColor(GamingTextColor);
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		g.drawString("LEVEL:", 0, gridYStart, Graphics.LEFT | Graphics.TOP);
		g.drawString(Integer.toString(level), gridXStart/2, gridYStart+20, Graphics.HCENTER | Graphics.TOP);
		g.drawString("SCORE:", 0, gridYStart + 60, Graphics.LEFT | Graphics.TOP);
		g.drawString(Long.toString(score), gridXStart/2, gridYStart+80, Graphics.HCENTER | Graphics.TOP);
		g.drawString("COUNT:", 0, gridYStart + 120, Graphics.LEFT | Graphics.TOP);
		g.drawString(Integer.toString(passedBlockCount), gridXStart/2, gridYStart+140, Graphics.HCENTER | Graphics.TOP);		
		//paint the next blocks
		currentBlock.paint(g, gridXEnd + 6, gridYStart, 48, 48);
		g.setColor(CurrentBlockBackgroundColor);
		g.drawRect(gridXEnd + 6, gridYStart, 48, 48);
		for(int i=0; i<nextBlocks.length; i++)
			nextBlocks[i].paint(g, gridXEnd + 6, gridYStart + (i+1)*50 + 2, 48, 48);
		//paint soft menu
		paintSoftMenu(g, "", "PAUSE");
	}

	private void paintSplash(Graphics g){
		g.setColor(0x00FFC0CB);
		g.fillRect(0, 0, getWidth(), getHeight());
		g.setColor(0x00000000);
		
		int lineGap = 22;
		int paraGap = 26;
		int titleGap = 30;
		int start = 12;
		
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_MEDIUM));
		g.drawString("Tetris", getWidth()/2, start, Graphics.HCENTER | Graphics.TOP);		
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		start += titleGap;
		
		g.drawString("Just a little game.", getWidth()/2, start, Graphics.HCENTER | Graphics.TOP); start += lineGap;
		g.drawString("Hope you like it!", getWidth()/2, start, Graphics.HCENTER | Graphics.TOP); start += lineGap;
	}
	
	private void paintSoftMenu(Graphics g, String left, String right){
		g.setColor(SoftBottonColor);
		g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
		g.drawString(left, 0, getHeight(), Graphics.LEFT | Graphics.BOTTOM);
		g.drawString(right, getWidth(), getHeight(), Graphics.RIGHT | Graphics.BOTTOM);
	}

	public void onBlockTimer() {
		if(status != Gaming) return;
		checkAndReduce();
		currentBlock.moveDown(grid);
		this.repaint();
	}

	public void checkAndReduce(){
		//check if touched the bottom
		boolean touchedBottom = false;
		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){				
				if(Block.Patterns[currentBlock.Type][currentBlock.Rotate][i][j] <= 0) continue;
				int x = currentBlock.StartX + i;
				int y = currentBlock.StartY + j + 1;
				if(x < 0 || x >= GridWidth || y >= GridHeight || (y >= 0 && grid[x][y] > 0))
					{touchedBottom = true; break;}
			}
			if(touchedBottom) break;
		}
		if(touchedBottom == false) return;
		//merge the block to grid
		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){				
				if(Block.Patterns[currentBlock.Type][currentBlock.Rotate][i][j] <= 0) continue;
				int x = currentBlock.StartX + i;
				int y = currentBlock.StartY + j;
				if(y < 0) {endGame();return;}
				grid[x][y] = 1;
				gridColors[x][y] = currentBlock.Color;
			}
		}
		//get new current block
		currentBlock = nextBlocks[0];
		for(int i=0; i<nextBlocks.length - 1; i++)
			nextBlocks[i] = nextBlocks[i+1];
		nextBlocks[nextBlocks.length - 1] = Block.CreateRandomBlock();
		passedBlockCount++;
		//reduce the full column rows
		for(int j=GridHeight-1; j>=0; ){
			boolean canReduce = true;
			for(int i=0; i<GridWidth; i++)
				if(grid[i][j] <= 0){canReduce = false; break;}
			if(canReduce == false) {j--; continue;}
			score += 10*level;
			for(int n=j; n>=1; n--)
				for(int m=0; m<GridWidth; m++)
					grid[m][n] = grid[m][n-1];
			for(int k=0; k<GridWidth; k++)
				grid[k][0] = 0;
		}
		//upgrade the level
		if((passedBlockCount/BlockCountPerLevel + 1) > level){
			if(level >= maxLevel) {endGame();return;}
			else {
				level++;
				blockTimer.cancel();
				blockTimer = new Timer();
				blockTimer.scheduleAtFixedRate(new BlockTimerTask(this), 0,	getBlockTimerTicksByLevel());
			}
		}
	}
}